@@ -7,9 +7,10 @@ from collections import deque
 from heapq import heappop, heappush
 from queue import Empty as SyncQueueEmpty
 from queue import Full as SyncQueueFull
+from time import monotonic
 from typing import Any, Callable, Generic, Optional, Protocol, TypeVar
 
-__version__ = "1.1.0"
+__version__ = "1.2.0"
 __all__ = (
     "Queue",
     "PriorityQueue",
@@ -30,62 +31,51 @@ OptFloat = Optional[float]
 
 class BaseQueue(Protocol[T]):
     @property
-    def maxsize(self) -> int:
-        ...
+    def maxsize(self) -> int: ...
 
     @property
-    def closed(self) -> bool:
-        ...
+    def closed(self) -> bool: ...
 
-    def task_done(self) -> None:
-        ...
+    def task_done(self) -> None: ...
 
-    def qsize(self) -> int:
-        ...
+    def qsize(self) -> int: ...
 
     @property
-    def unfinished_tasks(self) -> int:
-        ...
+    def unfinished_tasks(self) -> int: ...
 
-    def empty(self) -> bool:
-        ...
+    def empty(self) -> bool: ...
 
-    def full(self) -> bool:
-        ...
+    def full(self) -> bool: ...
 
-    def put_nowait(self, item: T) -> None:
-        ...
+    def put_nowait(self, item: T) -> None: ...
 
-    def get_nowait(self) -> T:
-        ...
+    def get_nowait(self) -> T: ...
 
 
 class SyncQueue(BaseQueue[T], Protocol[T]):
 
-    def put(self, item: T, block: bool = True, timeout: OptFloat = None) -> None:
-        ...
+    def put(self, item: T, block: bool = True, timeout: OptFloat = None) -> None: ...
 
-    def get(self, block: bool = True, timeout: OptFloat = None) -> T:
-        ...
+    def get(self, block: bool = True, timeout: OptFloat = None) -> T: ...
 
-    def join(self) -> None:
-        ...
+    def join(self) -> None: ...
 
 
 class AsyncQueue(BaseQueue[T], Protocol[T]):
-    async def put(self, item: T) -> None:
-        ...
+    async def put(self, item: T) -> None: ...
 
-    async def get(self) -> T:
-        ...
+    async def get(self) -> T: ...
 
-    async def join(self) -> None:
-        ...
+    async def join(self) -> None: ...
 
 
 class Queue(Generic[T]):
+    _loop: Optional[asyncio.AbstractEventLoop] = None
+
     def __init__(self, maxsize: int = 0) -> None:
-        self._loop = asyncio.get_running_loop()
+        if sys.version_info < (3, 10):
+            self._loop = asyncio.get_running_loop()
+
         self._maxsize = maxsize
 
         self._init(maxsize)
@@ -94,48 +84,51 @@ class Queue(Generic[T]):
 
         self._sync_mutex = threading.Lock()
         self._sync_not_empty = threading.Condition(self._sync_mutex)
+        self._sync_not_empty_waiting = 0
         self._sync_not_full = threading.Condition(self._sync_mutex)
-        self._all_tasks_done = threading.Condition(self._sync_mutex)
+        self._sync_not_full_waiting = 0
+        self._sync_tasks_done = threading.Condition(self._sync_mutex)
+        self._sync_tasks_done_waiting = 0
 
         self._async_mutex = asyncio.Lock()
         if sys.version_info[:3] == (3, 10, 0):
             # Workaround for Python 3.10 bug, see #358:
             getattr(self._async_mutex, "_get_loop", lambda: None)()
         self._async_not_empty = asyncio.Condition(self._async_mutex)
+        self._async_not_empty_waiting = 0
         self._async_not_full = asyncio.Condition(self._async_mutex)
-        self._finished = asyncio.Event()
-        self._finished.set()
+        self._async_not_full_waiting = 0
+        self._async_tasks_done = asyncio.Condition(self._async_mutex)
+        self._async_tasks_done_waiting = 0
 
         self._closing = False
-        self._pending: set[asyncio.Future[Any]] = set()
-
-        def checked_call_soon_threadsafe(
-            callback: Callable[..., None], *args: Any
-        ) -> None:
-            try:
-                self._loop.call_soon_threadsafe(callback, *args)
-            except RuntimeError:
-                # swallowing agreed in #2
-                pass
-
-        self._call_soon_threadsafe = checked_call_soon_threadsafe
-
-        def checked_call_soon(callback: Callable[..., None], *args: Any) -> None:
-            if not self._loop.is_closed():
-                self._loop.call_soon(callback, *args)
-
-        self._call_soon = checked_call_soon
+        self._pending: deque[asyncio.Future[Any]] = deque()
 
         self._sync_queue = _SyncQueueProxy(self)
         self._async_queue = _AsyncQueueProxy(self)
 
+    def _get_loop(self) -> asyncio.AbstractEventLoop:
+        # Warning!
+        # The function should be called when self._sync_mutex is locked,
+        # otherwise the code is not thread-safe
+        loop = asyncio.get_running_loop()
+
+        if self._loop is None:
+            self._loop = loop
+        if loop is not self._loop:
+            raise RuntimeError(f"{self!r} is bound to a different event loop")
+        return loop
+
     def close(self) -> None:
         with self._sync_mutex:
             self._closing = True
             for fut in self._pending:
                 fut.cancel()
-            self._finished.set()  # unblocks all async_q.join()
-            self._all_tasks_done.notify_all()  # unblocks all sync_q.join()
+            if self._async_tasks_done_waiting:
+                # unblocks all async_q.join()
+                self._notify_async(self._async_tasks_done.notify_all)
+            if self._sync_tasks_done_waiting:
+                self._sync_tasks_done.notify_all()  # unblocks all sync_q.join()
 
     async def wait_closed(self) -> None:
         # should be called from loop after close().
@@ -145,13 +138,17 @@ class Queue(Generic[T]):
             raise RuntimeError("Waiting for non-closed queue")
         # give execution chances for the task-done callbacks
         # of async tasks created inside
-        # _notify_async_not_empty, _notify_async_not_full
+        # _make_async_not_empty_notifier, _make_async_not_full_notifier
         # methods.
         await asyncio.sleep(0)
         if not self._pending:
             return
         await asyncio.wait(self._pending)
 
+    async def aclose(self) -> None:
+        self.close()
+        await self.wait_closed()
+
     @property
     def closed(self) -> bool:
         return self._closing and not self._pending
@@ -189,53 +186,27 @@ class Queue(Generic[T]):
     def _put_internal(self, item: T) -> None:
         self._put(item)
         self._unfinished_tasks += 1
-        self._finished.clear()
 
-    def _notify_sync_not_empty(self) -> None:
-        def f() -> None:
-            with self._sync_mutex:
-                self._sync_not_empty.notify()
+    async def _do_async_notifier(self, method: Callable[[], None]) -> None:
+        async with self._async_mutex:
+            method()
 
-        self._loop.run_in_executor(None, f)
+    def _setup_async_notifier(
+        self, loop: asyncio.AbstractEventLoop, method: Callable[[], None]
+    ) -> None:
+        task = loop.create_task(self._do_async_notifier(method))
+        task.add_done_callback(self._pending.remove)
+        self._pending.append(task)
 
-    def _notify_sync_not_full(self) -> None:
-        def f() -> None:
-            with self._sync_mutex:
-                self._sync_not_full.notify()
-
-        fut = asyncio.ensure_future(self._loop.run_in_executor(None, f))
-        fut.add_done_callback(self._pending.discard)
-        self._pending.add(fut)
-
-    def _notify_async_not_empty(self, *, threadsafe: bool) -> None:
-        async def f() -> None:
-            async with self._async_mutex:
-                self._async_not_empty.notify()
-
-        def task_maker() -> None:
-            task = self._loop.create_task(f())
-            task.add_done_callback(self._pending.discard)
-            self._pending.add(task)
-
-        if threadsafe:
-            self._call_soon_threadsafe(task_maker)
-        else:
-            self._call_soon(task_maker)
-
-    def _notify_async_not_full(self, *, threadsafe: bool) -> None:
-        async def f() -> None:
-            async with self._async_mutex:
-                self._async_not_full.notify()
-
-        def task_maker() -> None:
-            task = self._loop.create_task(f())
-            task.add_done_callback(self._pending.discard)
-            self._pending.add(task)
-
-        if threadsafe:
-            self._call_soon_threadsafe(task_maker)
-        else:
-            self._call_soon(task_maker)
+    def _notify_async(self, method: Callable[[], None]) -> None:
+        # Warning!
+        # The function should be called when self._sync_mutex is locked,
+        # otherwise the code is not thread-safe
+        loop = self._loop
+        if loop is None or loop.is_closed():
+            # async API is not available, nothing to notify
+            return
+        loop.call_soon_threadsafe(self._setup_async_notifier, loop, method)
 
     def _check_closing(self) -> None:
         if self._closing:
@@ -273,15 +244,18 @@ class _SyncQueueProxy(SyncQueue[T]):
         Raises a ValueError if called more times than there were items
         placed in the queue.
         """
-        self._parent._check_closing()
-        with self._parent._all_tasks_done:
-            unfinished = self._parent._unfinished_tasks - 1
+        parent = self._parent
+        parent._check_closing()
+        with parent._sync_tasks_done:
+            unfinished = parent._unfinished_tasks - 1
             if unfinished <= 0:
                 if unfinished < 0:
                     raise ValueError("task_done() called too many times")
-                self._parent._all_tasks_done.notify_all()
-                self._parent._loop.call_soon_threadsafe(self._parent._finished.set)
-            self._parent._unfinished_tasks = unfinished
+                if parent._sync_tasks_done_waiting:
+                    parent._sync_tasks_done.notify_all()
+                if parent._async_tasks_done_waiting:
+                    parent._notify_async(parent._async_tasks_done.notify_all)
+            parent._unfinished_tasks = unfinished
 
     def join(self) -> None:
         """Blocks until all items in the Queue have been gotten and processed.
@@ -292,11 +266,16 @@ class _SyncQueueProxy(SyncQueue[T]):
 
         When the count of unfinished tasks drops to zero, join() unblocks.
         """
-        self._parent._check_closing()
-        with self._parent._all_tasks_done:
-            while self._parent._unfinished_tasks:
-                self._parent._all_tasks_done.wait()
-                self._parent._check_closing()
+        parent = self._parent
+        parent._check_closing()
+        with parent._sync_tasks_done:
+            while parent._unfinished_tasks:
+                parent._sync_tasks_done_waiting += 1
+                try:
+                    parent._sync_tasks_done.wait()
+                finally:
+                    parent._sync_tasks_done_waiting -= 1
+                parent._check_closing()
 
     def qsize(self) -> int:
         """Return the approximate size of the queue (not reliable!)."""
@@ -328,7 +307,8 @@ class _SyncQueueProxy(SyncQueue[T]):
         condition where a queue can shrink before the result of full() or
         qsize() can be used.
         """
-        return 0 < self._parent._maxsize <= self._parent._qsize()
+        parent = self._parent
+        return 0 < parent._maxsize <= parent._qsize()
 
     def put(self, item: T, block: bool = True, timeout: OptFloat = None) -> None:
         """Put an item into the queue.
@@ -341,28 +321,38 @@ class _SyncQueueProxy(SyncQueue[T]):
         is immediately available, else raise the Full exception ('timeout'
         is ignored in that case).
         """
-        self._parent._check_closing()
-        with self._parent._sync_not_full:
-            if self._parent._maxsize > 0:
+        parent = self._parent
+        parent._check_closing()
+        with parent._sync_not_full:
+            if parent._maxsize > 0:
                 if not block:
-                    if self._parent._qsize() >= self._parent._maxsize:
+                    if parent._qsize() >= parent._maxsize:
                         raise SyncQueueFull
                 elif timeout is None:
-                    while self._parent._qsize() >= self._parent._maxsize:
-                        self._parent._sync_not_full.wait()
+                    while parent._qsize() >= parent._maxsize:
+                        parent._sync_not_full_waiting += 1
+                        try:
+                            parent._sync_not_full.wait()
+                        finally:
+                            parent._sync_not_full_waiting -= 1
                 elif timeout < 0:
                     raise ValueError("'timeout' must be a non-negative number")
                 else:
-                    time = self._parent._loop.time
-                    endtime = time() + timeout
-                    while self._parent._qsize() >= self._parent._maxsize:
-                        remaining = endtime - time()
+                    endtime = monotonic() + timeout
+                    while parent._qsize() >= parent._maxsize:
+                        remaining = endtime - monotonic()
                         if remaining <= 0.0:
                             raise SyncQueueFull
-                        self._parent._sync_not_full.wait(remaining)
-            self._parent._put_internal(item)
-            self._parent._sync_not_empty.notify()
-            self._parent._notify_async_not_empty(threadsafe=True)
+                        parent._sync_not_full_waiting += 1
+                        try:
+                            parent._sync_not_full.wait(remaining)
+                        finally:
+                            parent._sync_not_full_waiting -= 1
+            parent._put_internal(item)
+            if parent._sync_not_empty_waiting:
+                parent._sync_not_empty.notify()
+            if parent._async_not_empty_waiting:
+                parent._notify_async(parent._async_not_empty.notify)
 
     def get(self, block: bool = True, timeout: OptFloat = None) -> T:
         """Remove and return an item from the queue.
@@ -375,27 +365,37 @@ class _SyncQueueProxy(SyncQueue[T]):
         available, else raise the Empty exception ('timeout' is ignored
         in that case).
         """
-        self._parent._check_closing()
-        with self._parent._sync_not_empty:
+        parent = self._parent
+        parent._check_closing()
+        with parent._sync_not_empty:
             if not block:
-                if not self._parent._qsize():
+                if not parent._qsize():
                     raise SyncQueueEmpty
             elif timeout is None:
-                while not self._parent._qsize():
-                    self._parent._sync_not_empty.wait()
+                while not parent._qsize():
+                    parent._sync_not_empty_waiting += 1
+                    try:
+                        parent._sync_not_empty.wait()
+                    finally:
+                        parent._sync_not_empty_waiting -= 1
             elif timeout < 0:
                 raise ValueError("'timeout' must be a non-negative number")
             else:
-                time = self._parent._loop.time
-                endtime = time() + timeout
-                while not self._parent._qsize():
-                    remaining = endtime - time()
+                endtime = monotonic() + timeout
+                while not parent._qsize():
+                    remaining = endtime - monotonic()
                     if remaining <= 0.0:
                         raise SyncQueueEmpty
-                    self._parent._sync_not_empty.wait(remaining)
-            item = self._parent._get()
-            self._parent._sync_not_full.notify()
-            self._parent._notify_async_not_full(threadsafe=True)
+                    parent._sync_not_empty_waiting += 1
+                    try:
+                        parent._sync_not_empty.wait(remaining)
+                    finally:
+                        parent._sync_not_empty_waiting -= 1
+            item = parent._get()
+            if parent._sync_not_full_waiting:
+                parent._sync_not_full.notify()
+            if parent._async_not_full_waiting:
+                parent._notify_async(parent._async_not_full.notify)
             return item
 
     def put_nowait(self, item: T) -> None:
@@ -426,21 +426,25 @@ class _AsyncQueueProxy(AsyncQueue[T]):
 
     @property
     def closed(self) -> bool:
-        return self._parent.closed
+        parent = self._parent
+        return parent.closed
 
     def qsize(self) -> int:
         """Number of items in the queue."""
-        return self._parent._qsize()
+        parent = self._parent
+        return parent._qsize()
 
     @property
     def unfinished_tasks(self) -> int:
         """Return the number of unfinished tasks."""
-        return self._parent._unfinished_tasks
+        parent = self._parent
+        return parent._unfinished_tasks
 
     @property
     def maxsize(self) -> int:
         """Number of items allowed in the queue."""
-        return self._parent._maxsize
+        parent = self._parent
+        return parent._maxsize
 
     def empty(self) -> bool:
         """Return True if the queue is empty, False otherwise."""
@@ -452,10 +456,11 @@ class _AsyncQueueProxy(AsyncQueue[T]):
         Note: if the Queue was initialized with maxsize=0 (the default),
         then full() is never True.
         """
-        if self._parent._maxsize <= 0:
+        parent = self._parent
+        if parent._maxsize <= 0:
             return False
         else:
-            return self.qsize() >= self._parent._maxsize
+            return parent._qsize() >= parent._maxsize
 
     async def put(self, item: T) -> None:
         """Put an item into the queue.
@@ -465,43 +470,43 @@ class _AsyncQueueProxy(AsyncQueue[T]):
 
         This method is a coroutine.
         """
-        self._parent._check_closing()
-        async with self._parent._async_not_full:
-            self._parent._sync_mutex.acquire()
-            locked = True
-            try:
-                if self._parent._maxsize > 0:
-                    do_wait = True
-                    while do_wait:
-                        do_wait = self._parent._qsize() >= self._parent._maxsize
-                        if do_wait:
-                            locked = False
-                            self._parent._sync_mutex.release()
-                            await self._parent._async_not_full.wait()
-                            self._parent._sync_mutex.acquire()
-                            locked = True
+        parent = self._parent
+        parent._check_closing()
+        async with parent._async_not_full:
+            with parent._sync_mutex:
+                parent._get_loop()  # check the event loop
+                while 0 < parent._maxsize <= parent._qsize():
+                    parent._async_not_full_waiting += 1
+                    parent._sync_mutex.release()
+                    try:
+                        await parent._async_not_full.wait()
+                    finally:
+                        parent._sync_mutex.acquire()
+                        parent._async_not_full_waiting -= 1
 
-                self._parent._put_internal(item)
-                self._parent._async_not_empty.notify()
-                self._parent._notify_sync_not_empty()
-            finally:
-                if locked:
-                    self._parent._sync_mutex.release()
+                parent._put_internal(item)
+                if parent._async_not_empty_waiting:
+                    parent._async_not_empty.notify()
+                if parent._sync_not_empty_waiting:
+                    parent._sync_not_empty.notify()
 
     def put_nowait(self, item: T) -> None:
         """Put an item into the queue without blocking.
 
         If no free slot is immediately available, raise QueueFull.
         """
-        self._parent._check_closing()
-        with self._parent._sync_mutex:
-            if self._parent._maxsize > 0:
-                if self._parent._qsize() >= self._parent._maxsize:
-                    raise AsyncQueueFull
+        parent = self._parent
+        parent._check_closing()
+        with parent._sync_mutex:
+            parent._get_loop()
+            if 0 < parent._maxsize <= parent._qsize():
+                raise AsyncQueueFull
 
-            self._parent._put_internal(item)
-            self._parent._notify_async_not_empty(threadsafe=False)
-            self._parent._notify_sync_not_empty()
+            parent._put_internal(item)
+            if parent._async_not_empty_waiting:
+                parent._notify_async(parent._async_not_empty.notify)
+            if parent._sync_not_empty_waiting:
+                parent._sync_not_empty.notify()
 
     async def get(self) -> T:
         """Remove and return an item from the queue.
@@ -510,43 +515,44 @@ class _AsyncQueueProxy(AsyncQueue[T]):
 
         This method is a coroutine.
         """
-        self._parent._check_closing()
-        async with self._parent._async_not_empty:
-            self._parent._sync_mutex.acquire()
-            locked = True
-            try:
-                do_wait = True
-                while do_wait:
-                    do_wait = self._parent._qsize() == 0
+        parent = self._parent
+        parent._check_closing()
+        async with parent._async_not_empty:
+            with parent._sync_mutex:
+                parent._get_loop()  # check the event loop
+                while not parent._qsize():
+                    parent._async_not_empty_waiting += 1
+                    parent._sync_mutex.release()
+                    try:
+                        await parent._async_not_empty.wait()
+                    finally:
+                        parent._sync_mutex.acquire()
+                        parent._async_not_empty_waiting -= 1
 
-                    if do_wait:
-                        locked = False
-                        self._parent._sync_mutex.release()
-                        await self._parent._async_not_empty.wait()
-                        self._parent._sync_mutex.acquire()
-                        locked = True
-
-                item = self._parent._get()
-                self._parent._async_not_full.notify()
-                self._parent._notify_sync_not_full()
+                item = parent._get()
+                if parent._async_not_full_waiting:
+                    parent._async_not_full.notify()
+                if parent._sync_not_full_waiting:
+                    parent._sync_not_full.notify()
                 return item
-            finally:
-                if locked:
-                    self._parent._sync_mutex.release()
 
     def get_nowait(self) -> T:
         """Remove and return an item from the queue.
 
         Return an item if one is immediately available, else raise QueueEmpty.
         """
-        self._parent._check_closing()
-        with self._parent._sync_mutex:
-            if self._parent._qsize() == 0:
+        parent = self._parent
+        parent._check_closing()
+        with parent._sync_mutex:
+            if not parent._qsize():
                 raise AsyncQueueEmpty
 
-            item = self._parent._get()
-            self._parent._notify_async_not_full(threadsafe=False)
-            self._parent._notify_sync_not_full()
+            parent._get_loop()
+            item = parent._get()
+            if parent._async_not_full_waiting:
+                parent._notify_async(parent._async_not_full.notify)
+            if parent._sync_not_full_waiting:
+                parent._sync_not_full.notify()
             return item
 
     def task_done(self) -> None:
@@ -563,14 +569,17 @@ class _AsyncQueueProxy(AsyncQueue[T]):
         Raises ValueError if called more times than there were items placed in
         the queue.
         """
-        self._parent._check_closing()
-        with self._parent._all_tasks_done:
-            if self._parent._unfinished_tasks <= 0:
+        parent = self._parent
+        parent._check_closing()
+        with parent._sync_tasks_done:
+            if parent._unfinished_tasks <= 0:
                 raise ValueError("task_done() called too many times")
-            self._parent._unfinished_tasks -= 1
-            if self._parent._unfinished_tasks == 0:
-                self._parent._finished.set()
-                self._parent._all_tasks_done.notify_all()
+            parent._unfinished_tasks -= 1
+            if parent._unfinished_tasks == 0:
+                if parent._async_tasks_done_waiting:
+                    parent._notify_async(parent._async_tasks_done.notify_all)
+                if parent._sync_tasks_done_waiting:
+                    parent._sync_tasks_done.notify_all()
 
     async def join(self) -> None:
         """Block until all items in the queue have been gotten and processed.
@@ -580,12 +589,20 @@ class _AsyncQueueProxy(AsyncQueue[T]):
         indicate that the item was retrieved and all work on it is complete.
         When the count of unfinished tasks drops to zero, join() unblocks.
         """
-        while True:
-            with self._parent._sync_mutex:
-                self._parent._check_closing()
-                if self._parent._unfinished_tasks == 0:
-                    break
-            await self._parent._finished.wait()
+        parent = self._parent
+        parent._check_closing()
+        async with parent._async_tasks_done:
+            with parent._sync_mutex:
+                parent._get_loop()  # check the event loop
+                while parent._unfinished_tasks:
+                    parent._async_tasks_done_waiting += 1
+                    parent._sync_mutex.release()
+                    try:
+                        await parent._async_tasks_done.wait()
+                    finally:
+                        parent._sync_mutex.acquire()
+                        parent._async_tasks_done_waiting -= 1
+                    parent._check_closing()
 
 
 class PriorityQueue(Queue[T]):
